/*
 *  templatelistview.cpp  -  widget showing list of alarm templates
 *  Program:  kalarm
 *  Copyright © 2004,2005,2007,2009 by David Jarvie <software@astrojar.org.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "kalarm.h"

#include <qpainter.h>
#include <qpixmap.h>
#include <klocale.h>
#include <kdebug.h>

#include "alarmcalendar.h"
#include "alarmresource.h"
#include "functions.h"
#include "templatelistview.moc"


/*=============================================================================
=  Class: TemplateListView
=  Displays the list of outstanding alarms.
=============================================================================*/
QValueList<EventListViewBase*>  TemplateListView::mInstanceList;


TemplateListView::TemplateListView(Type show, Type enable, const QString& whatsThisText, QWidget* parent, const char* name)
	: EventListViewBase(KCalEvent::TEMPLATE, parent, name),
	  mWhatsThisText(whatsThisText),
	  mIconColumn(0),
	  mNameColumn(1),
	  mTypesEnabled(enable),
	  mTypeFilter(show)
{
	addColumn(QString::null);          // icon column
	addLastColumn(i18n("Name"));
	setSorting(mNameColumn);           // sort initially by name
	// Set the width of the alarm type column to exactly accommodate the icons.
	setColumnWidth(mIconColumn, EventListViewItemBase::typeIconWidth(this));
	setColumnWidthMode(mIconColumn, QListView::Manual);
	setColumnAlignment(mIconColumn, Qt::AlignHCenter);

	mInstanceList.append(this);
}

TemplateListView::~TemplateListView()
{
	mInstanceList.remove(this);
}

/******************************************************************************
*  Add all the templates to the list.
*/
void TemplateListView::populate()
{
	KAEvent::List templates = KAlarm::templateList();
	for (KAEvent::List::Iterator it = templates.begin();  it != templates.end();  ++it)
	{
		Type type;
		switch ((*it)->action())
		{
			case KAEvent::MESSAGE:
			case KAEvent::FILE:     type = DISPLAY;  break;
			case KAEvent::COMMAND:  type = COMMAND;  break;
			case KAEvent::EMAIL:    type = EMAIL;  break;
			default:                type = ALL;  break;
		}
		if (type & mTypeFilter)
			addEntry(*it);
	}
}

/******************************************************************************
*  Create a new list item for addEntry().
*/
EventListViewItemBase* TemplateListView::createItem(KAEvent* event)
{
	return new TemplateListViewItem(mTypesEnabled, this, event);
}

void TemplateListView::setSelected(TemplateListViewItem* item, bool selected)
{
	EventListViewBase::setSelected(item, selected);
}

/******************************************************************************
*  Returns the QWhatsThis text for a specified column.
*/
QString TemplateListView::whatsThisText(int column) const
{
	if (column == mIconColumn)
		return i18n("Alarm type");
	if (column == mNameColumn)
		return i18n("Name of the alarm template");
	return mWhatsThisText;
}


/*=============================================================================
=  Class: TemplateListViewItem
=  Contains the details of one alarm for display in the TemplateListView.
=============================================================================*/

TemplateListViewItem::TemplateListViewItem(TemplateListView::Type type, TemplateListView* parent, KAEvent* event)
	: EventListViewItemBase(parent, event)
{
	setLastColumnText();     // set the template name column text

	int index;
	switch (event->action())
	{
		case KAAlarm::FILE:
			index = 2;
			mEnabled = (type & TemplateListView::DISPLAY);
			break;
		case KAAlarm::COMMAND:
			index = 3;
			mEnabled = (type & TemplateListView::COMMAND);
			break;
		case KAAlarm::EMAIL:
			index = 4;
			mEnabled = (type & TemplateListView::EMAIL);
			break;
		case KAAlarm::MESSAGE:
		default:
			index = 1;
			mEnabled = (type & TemplateListView::DISPLAY);
			break;
	}
kdDebug()<<"TemplateListViewItem(): type="<<type<<", enabled="<<mEnabled<<endl;
	setSelectable(mEnabled);
	mIconOrder.sprintf("%01u", index);
	setPixmap(templateListView()->iconColumn(), *eventIcon());
}

/******************************************************************************
*  Return the alarm summary text.
*/
QString TemplateListViewItem::lastColumnText() const
{
	return event()->templateName();
}

/******************************************************************************
*  Return the column sort order for one item in the list.
*/
QString TemplateListViewItem::key(int column, bool) const
{
	TemplateListView* listView = templateListView();
	if (column == listView->iconColumn())
		return mIconOrder;
	return text(column).lower();
}

/******************************************************************************
*  Paint one value in one of the columns in the list view.
*/
void TemplateListViewItem::paintCell(QPainter* painter, const QColorGroup& cgrp, int column, int width, int /*align*/)
{
	const TemplateListView* listView = templateListView();
	int    margin = listView->itemMargin();
	QRect  box(margin, margin, width - margin*2, height() - margin*2);   // area within which to draw
	bool   selected = isSelected();
	const QColorGroup& cg = mEnabled ? cgrp : templateListView()->palette().disabled();
	QColor bgColour = selected ? cg.highlight()
	                : (resource() && resource()->colour().isValid()) ? resource()->colour()
	                : cg.base();
	QColor fgColour = selected ? cg.highlightedText() : cg.text();
	painter->setPen(fgColour);
	painter->fillRect(0, 0, width, height(), bgColour);
	if (column == listView->iconColumn())
	{
		// Display the alarm type icon, horizontally and vertically centred in the cell
		QPixmap* pixmap = eventIcon();
		QRect pixmapRect = pixmap->rect();
		int diff = box.height() - pixmap->height();
		if (diff < 0)
		{
			pixmapRect.setTop(-diff / 2);
			pixmapRect.setHeight(box.height());
		}
		QPoint iconTopLeft(box.left() + (box.width() - pixmapRect.width()) / 2,
		                   box.top() + (diff > 0 ? diff / 2 : 0));
		painter->drawPixmap(iconTopLeft, *pixmap, pixmapRect);
	}
	else if (column == listView->nameColumn())
		painter->drawText(box, AlignVCenter, text(column));
}

